/* Diseño de un medidor IoT doméstico 

 * Este código permite realizar una conexión bidireccional entre el dispositivo NodeMCU V3 y la plataforma IoT ThingSpeak,
 * Pudiendo enviar datos de temperatura y humedad obtenidos por el sensor DHT22, 
 * y leer datos enviados desde ThingSpeak para encerder o apagar un LED.
 * 
 * Para realizar dicha conexión se han implementado 2 metodos:
 *  1. Tpeak: Utiliza las funciones de la librería ThingSpeak.h para realizar la conexion y el intercambio de datos.
 *  2. MQTT: Utiliza el método publicador-subscriptor basado en el protocolo MQTT para relizar la conexion y el intercambio de datos.
 * 
 */

#include <DHT.h> //Libreria de sensores DHT
#include <ESP8266WiFi.h> // Para realizar la conexión a internet
#include <ThingSpeak.h> //Librería plataforma IoT
#include <PubSubClient.h> //Método publicador-subscriptor

#define DHTPIN D5 //Pin al que se conecta el DHT22
#define DHTTYPE DHT22 //Se define el tipo de sensor
#define LED D1 //Pin al que se conecta el LED 

//Se definen los dos métodos para realizar la conexión con la plataforma
#define Tpeak // Utilizando la libreria de la plataforma IoT ThingSpeak
//#define MQTT //Utilizando el protocolo MQTT


// Credenciales para la conexion wifi
char* ssid = "************"; // Nombre del WiFi
char* pass = "******"; // Contraseña

// Datos necesarios de Thingspeak para realizar la conexion del NodeMCU con la plataforma
unsigned long myChannelNumber1 = 1224024; // Thingspeak número de canal
char* myWriteAPIKey1 = "ID5A1VDZEGPVDWIW"; // ThingSpeak write API Key
char* myReadAPIKey1 = "UV02Q6HUBRSKA85M"; // ThingSpeak read API Key

unsigned long myChannelNumber2 = 1232890; // Thingspeak número de canal
char* myWriteAPIKey2 = "MZG4QNQWVG6P34DI"; // ThingSpeak write API Key
char* myReadAPIKey2 = "FXTQY1CD8JH4CHFP"; // ThingSpeak read API Key  

char* MQTTAPIKey = "ZRRRYXR2L7Z07C7L"; //Clave MQTT de thingspeak
char* clientName = "ESP-Thingspeak";

const unsigned int fieldLED = 1; //campo 1 del canal 2 del que leemos para cambiar el estado del led
const unsigned int fieldt = 1; // campo 1 del canal 1 al que enviamos la temperatura
const unsigned int fieldh = 2; // campo 2 del canal 1 al que enviamos la humedad


char* server = "mqtt.thingspeak.com"; // Servidor o Broker de la plataforma ThingSpeak



DHT dht(DHTPIN, DHTTYPE); // Se crea el objeto DHT 

WiFiClient client; // Se define el tipo de cliente 

PubSubClient mqttClient( client ); //Se crea el objeto cliente

int status = WL_IDLE_STATUS; // Variable para deterinar el estado de conexion WiFi


void setup() 
{

  Serial.begin(9600); // Velocidad puerto serie

  delay(10);

  dht.begin(); // Inicia el sensor

  pinMode(LED, OUTPUT); // Se configura el pin como salida

  WiFiConnect(); //Se realiza la conexión WiFi


  #ifdef Tpeak

    ThingSpeak.begin(client); // Inicia ThingSpeak

  #else

    mqttClient.setServer( server, 1883 ); // Se establecen los detalles del servidor MQTT.
    
    mqttClient.setCallback(myMessageArrived); // Detecta eventos en el canal de lectura 

  #endif

}


void loop() 
{

  float dataSensor[2]; // vector que almacena los valores de temperatura y humedad


  #ifdef Tpeak

  // Se leen temperatura y humedad
    readSensor(dataSensor);

  // Se envían temperatura y humedad a ThingSpeak
    sendingData(myChannelNumber1, myWriteAPIKey1, fieldt, fieldh, dataSensor);

  // Se lee del canal de ThingSpeak
    readingData(myChannelNumber2, fieldLED, myReadAPIKey2);

  // Se repite el proceso cada minuto
    delay(15000);

  #else

    // Si el cliente MQTT no está conectado se realiza la conexion 
    if (!mqttClient.connected())
    {
     mqttConnect(); 

     // Se realiza la subscripcion al canal de ThingSpeak del que se van a leer los datos
      if (mqttSubscribe(myChannelNumber2, fieldLED, myReadAPIKey2) == 1 )
      {
        Serial.println( "Subscrito \n" );

      }
    }

    mqttClient.loop(); // Llamada al bucle para mantener la conexión con el servidor


   // Le leen temperatura y humedad
    readSensor(dataSensor);

  // Se envian datos a ThingSpeak
    mqttPublish(myChannelNumber1, myWriteAPIKey1, dataSensor);
    
  // Se repite el proceso cada 14 segundos
    delay(14000);

  #endif

}


//Funcion que se encarga de realizar la conexion wifi

void WiFiConnect()
{
  WiFi.begin(ssid, pass); // Inicia WiFi

  Serial.println();
  Serial.print("Conectando a ");
  Serial.print(ssid);

  while (WiFi.status() != WL_CONNECTED)
  {
    delay(500);
    Serial.print(".");
  }

  Serial.println();
  Serial.println("Conectado a WiFi");
  Serial.print("Dirección IP: ");
  Serial.println(WiFi.localIP());
  Serial.print("\n");

}


// Funcion encargada de leer la temperatura y la humedad con el sensor DHT22 y almacenar los datos en el vector data[]

void readSensor(float data[])
{
  float t;
  float h;
  
  // Se lee temperatura
  t = dht.readTemperature();
    
  // Se lee humedad
  h = dht.readHumidity();
  
  // Se comprueba si las lecturas pueden dar algún fallo mediante la función isnan()
  // Esta función devuelve un 1 en caso de que el valor no sea numérico
  while (isnan(t) || isnan(h)) {
    Serial.println("Fallo en la lectura del sensor DHT!");
    delay(1000);
    t = dht.readTemperature(); 
    h = dht.readHumidity();   
  }
  
  // Se asignan los valores de temperatura y humedad al vector
  data[0] = t;
  data[1] = h;

  Serial.print("Temperatura: ");
  Serial.print(t);
  Serial.print(" ºC\t ");
  
  Serial.print("Humedad: ");
  Serial.print(h);
  Serial.print(" %\t");
}
